import processing.core.*; 
import processing.data.*; 
import processing.opengl.*; 

import oscP5.*; 
import netP5.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class ga extends PApplet {

/*
  GA-1 / emoc / oct. 2010 -> apr. 2013 / processing 2.0b6 
  a genetic workbench to experiment with sounds 
  
  
  **********************************************************************
  
  Copyright (C) <2013>  <Pierre Commenge>
  P. Commenge - http://emoc.org
  oscP5 library by Andreas Schlegel - http://www.sojamo.de/

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
  **********************************************************************
  
  16 parameters are sent by OSC, each normalized from 0 to 1
  10 modes to switch
  
  OSC messages
  /mode i
  /params f f f f f f f f f f f f f f f f 
  
*/





OscP5 osca;
NetAddress loc;
int OscOutPort = 57120;                         // <-- OSC OUT PORT

String SKETCH_NAME = "ga";
PFont font;

ArrayList zorgs = new ArrayList();
boolean init = true;
Orga parentA = new Orga(205, 75, false);
Orga parentB = new Orga(335, 75, false);
int zorg_selected = -1;
int max_params = 16;

int mode = 5;
boolean help_visible = false;

SoundBank sound_bank = new SoundBank(550, 150);
Timeline timeline = new Timeline(50, 670);
boolean timeline_is_playing = false;

int[] color_modes = new int[10];
boolean redraw_colorbar = true;
boolean redraw_background = true;

public void setup() {
  size(950, 700);
  smooth();
  colorMode(HSB, 1.0f);
  color_modes[0] = color(0.0f, 1.0f, 1.0f, 1.0f);
  color_modes[1]=color(0.1f, 1.0f, 1.0f, 1.0f);
  color_modes[2]=color(0.2f, 1.0f, 1.0f, 1.0f);
  color_modes[3]=color(0.3f, 1.0f, 1.0f, 1.0f);
  color_modes[4]=color(0.5f, 1.0f, 1.0f, 1.0f);
  color_modes[5]=color(0.55f, 1.0f, 1.0f, 1.0f);
  color_modes[6]=color(0.6f, 1.0f, 1.0f, 1.0f);
  color_modes[7]=color(0.7f, 1.0f, 1.0f, 1.0f);
  color_modes[8]=color(0.8f, 1.0f, 1.0f, 1.0f);
  color_modes[9]=color(0.9f, 1.0f, 1.0f, 1.0f);
  font = loadFont("OhLaLa-10.vlw");
  osca = new OscP5(this, 12000);
  loc = new NetAddress("127.0.0.1", OscOutPort);
  frameRate(5);
  
}



public void draw() {
  if (redraw_background) {
    background(1);
    redraw_background = false;
  }
  textFont(font, 10);
  
  if (init) {
    for (int i = 0; i < 4; i++) {
      for (int j = 0; j < 3; j++) {
        float _x = 130*i + 75;
        float _y = 130*j + 75 + 130;
        zorgs.add( new Orga(_x, _y, true));
      }
    }
    sound_bank.loadBank();
    timeline.initBySoundBank(sound_bank);
    init = false;
  }
  
  if (timeline_is_playing) {
    if (timeline.is_playing == false) { // not already playing
      timeline.start(millis());
    } else {
      if (timeline.endReached()) timeline_is_playing = false;
      else timeline.play(millis());
    }
  }
  
  // afficher la banque de sons
  sound_bank.drawBank();
  
  // afficher tout
  zorg_selected = -1;
  for (int i = 0; i < 12; i++) {
    Orga z = (Orga) zorgs.get(i);
    if (z.testMouse(PApplet.parseFloat(mouseX), PApplet.parseFloat(mouseY))) zorg_selected = i;
    z.drawOrga();
  }
  
  // afficher les parents
  parentA.drawOrga();
  parentB.drawOrga();
  
  // afficher les parametres de la s\u00e9lection
  if (zorg_selected > -1) {
    //println("selected : " + zorg_selected);
    Orga z = (Orga) zorgs.get(zorg_selected);
    z.showParams();
  }
  
  if (redraw_colorbar) {
    drawColorBar();
    redraw_colorbar = false;
  }
  timeline.doActions();
  
  showMode();
  if (help_visible) showHelp();
}








public void showMode() {
  fill(1);
  rect(650, 48, 100, 15);
  fill(0, 0, 0);
  text("mode : " + mode, 650, 60);
}

public int getModeColor(int _mode) {
  return color_modes[_mode];
}

public void drawColorBar() {
  fill(1); rect(650, 0, 300, 40);
  float x_base = 650;
  stroke(0,0,1,1);
  for (int i = 0; i < 10; i++) { 
    fill(color_modes[i]); 
    rect(x_base + (i * 30), 0, 30, 30);
  }
  fill(0,0,0,1);
  rect(x_base + ((mode) * 30), 30, 30, 10);
}

public void breedParents() {
  /* 30% de r\u00e9cup\u00e9rer un g\u00e8ne du parentA
     30% de r\u00e9cup\u00e9rer un g\u00e8ne du parentB
     40% de r\u00e9cup\u00e9rer une moyenne des g\u00e8nes */
  if (parentA.alive && parentB.alive) {
    float[] new_genotype = new float[max_params];
    for (int i = 0; i < 12; i++) {
      Orga z = (Orga) zorgs.get(i);
      for (int j = 0; j < max_params; j++) {
        float rr = random(1);
        if (rr <= 0.3f) {
          new_genotype[j] = parentA.genotype[j];
        } else if ((rr > 0.3f) && (rr <= 0.6f)) {
          new_genotype[j] = parentB.genotype[j];
        } else {
          new_genotype[j] = mixGenes(parentA.genotype[j], parentB.genotype[j]);
        }
      }
      z.setGenotype(new_genotype); 
    }  
  }
}

public float mixGenes(float va, float vb) {
  float vresult;
  if (abs(va - vb) < 0.5f) {
    vresult = (va + vb) / 2;  
  } else {
    vresult = ((((1 - vb) + va) / 2) + vb) % 1;
    //println("cas 2, va :  " + va + ", vb : " + vb + ", vresult : " + vresult); 
  }
  return vresult;
} 


public void sendGenotypeOsc(float[] _genotype, int mode) {
  sendModeOsc(mode);
  OscMessage message = new OscMessage("/params");
  String debug_osc = "";
  for (int i = 0; i < max_params; i++) {
    message.add(_genotype[i]); 
    debug_osc = debug_osc + " " + _genotype[i];
  }
  println("message \u00e0 envoyer : " + debug_osc);
  osca.send(message, loc); 
}

public void sendModeOsc(int mode) {
  OscMessage message = new OscMessage("/mode");
  message.add(mode); 
  osca.send(message, loc); 
}

public void switchHelp() {
  help_visible = !help_visible;
  if (help_visible == false) redrawAll();
}

public void redrawAll() {
  redraw_background = true;
  redraw_colorbar = true;
  timeline.redraw_sounds = true;
  timeline.redraw_timeline = true;
  sound_bank.redraw_bank = true;
  for (int i = 0; i < 12; i++) {
    Orga z = (Orga) zorgs.get(i);
    z.redraw_me = true;
  }
}

/* **********************************************************************

                              CONTROLS
                         
   ********************************************************************** */
   
   
public void keyPressed() {

  if (key == 's') {
    saveFrame(SKETCH_NAME+"_"+year()+month()+day()+hour()+minute()+second()+millis()+".png");
  }
  if (key == 'c') {
    breedParents();
  }
  if (key == 'h') {
   switchHelp();
  }
  if (key == 'x') {
    sound_bank.saveBank();
  }
  if (key == 'y') {
    sound_bank.loadBank();
  }
  if (key == 'l') {
    timeline_is_playing = !timeline_is_playing;
  }
  if (key == 't') {
    timeline.saveTimeline();
  }
  if (key == 'r') {
    if (zorg_selected > -1) {
      Orga z = (Orga) zorgs.get(zorg_selected);
      z.randomGenotype();
    }
  }
  if (key == 'm') {
    if (zorg_selected > -1) {
      Orga z = (Orga) zorgs.get(zorg_selected);
      z.mutateGenotype(0.05f);
    }
  }
  if (key == 'n') {
    timeline_is_playing = false;
    //timeline.endReached();
    //timeline.initBySoundBank(sound_bank);
    timeline.reInit(sound_bank);
    //timeline.redraw_sounds = true;
  }  
  if (key == 'o') {
    if (zorg_selected > -1) {
      Orga z = (Orga) zorgs.get(zorg_selected);
      float[] gen = new float[max_params];
      gen = z.getGenotype();
      int zmode = z.mode;
      if (zmode == 0) zmode = mode;
      sendGenotypeOsc(gen, zmode);
    } else {
      int sbmin = sound_bank.mouseIn(PApplet.parseFloat(mouseX), PApplet.parseFloat(mouseY));
      if ( sbmin > -1) {
        println("mouse_in" + sound_bank.mouseIn(PApplet.parseFloat(mouseX), PApplet.parseFloat(mouseY)));
        // play from bank
        float[] gen = new float[max_params];
        Orga z = (Orga) sound_bank.sounds.get(sbmin);
        gen = z.getGenotype();
        int zmode = z.mode;
        print("envoi genotype depuis la banque");
        sendGenotypeOsc(gen, zmode);
      }
    }
  }
  if (key == 'd') {
    int sbmin = sound_bank.mouseIn(PApplet.parseFloat(mouseX), PApplet.parseFloat(mouseY));
    if ( sbmin > -1) {
      println("supprimer" + sbmin);
      //sound_bank.sounds.remove(sbmin);
      sound_bank.removeSound(sbmin);
      sound_bank.redraw_bank = true;
    }
  }
  
  if (key == 'a') {
    if (parentA.alive) {
      //Orga z = (Orga) zorgs.get(zorg_selected);
      float[] gen = new float[max_params];
      gen = parentA.getGenotype();
      sendGenotypeOsc(gen, mode);
    }
  }
  if (key == 'b') {
    if (parentB.alive) {
      //Orga z = (Orga) zorgs.get(zorg_selected);
      float[] gen = new float[max_params];
      gen = parentB.getGenotype();
      sendGenotypeOsc(gen, mode);
    }
  }
  
  if (key == 'k') {
    if (zorg_selected > -1) {
      Orga z = (Orga) zorgs.get(zorg_selected);
      float[] gen = new float[max_params];
      gen = z.getGenotype();
      int zmode = z.mode;
      if (zmode == 0) zmode = mode;
      //sendGenotypeOsc(gen, zmode);
      sound_bank.addSound(gen, zmode);
      sound_bank.redraw_bank = true;
    }
  }
  
  if (key == '0') { mode = 0; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '1') { mode = 1; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '2') { mode = 2; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '3') { mode = 3; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '4') { mode = 4; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '5') { mode = 5; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '6') { mode = 6; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '7') { mode = 7; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '8') { mode = 8; sendModeOsc(mode); redraw_colorbar = true; } 
  if (key == '9') { mode = 9; sendModeOsc(mode); redraw_colorbar = true; } 
}

public void mousePressed() {
  if (mouseButton == LEFT) {
    if (zorg_selected > -1) {
      Orga z = (Orga) zorgs.get(zorg_selected);
      for (int i = 0; i < max_params; i++) {
        parentA.genotype[i] = z.genotype[i];
      }
      parentA.alive = true;
    } else {
      int sbmin = sound_bank.mouseIn(PApplet.parseFloat(mouseX), PApplet.parseFloat(mouseY));
      if ( sbmin > -1) {
        Orga z = (Orga) sound_bank.sounds.get(sbmin);
        for (int i = 0; i < max_params; i++) {
          parentA.genotype[i] = z.genotype[i];
        }
        parentA.alive = true;
      }
    }
  } else if (mouseButton == RIGHT) {
    if (zorg_selected > -1) {
      Orga z = (Orga) zorgs.get(zorg_selected);
      for (int i = 0; i < max_params; i++) {
        parentB.genotype[i] = z.genotype[i];
      }
      parentB.alive = true;
    } else {
    int sbmin = sound_bank.mouseIn(PApplet.parseFloat(mouseX), PApplet.parseFloat(mouseY));
      if ( sbmin > -1) {
        Orga z = (Orga) sound_bank.sounds.get(sbmin);
        for (int i = 0; i < max_params; i++) {
          parentB.genotype[i] = z.genotype[i];
        }
        parentB.alive = true;
      }
    }
  }
}

class Orga {
  
  float[] genotype = new float[16]; // genotype, 16 parametres
  int mode = 0; // le mode g\u00e9n\u00e9ral (0) par d\u00e9faut
  float diametre_max = 112;
  float x, y;
  boolean alive;
  boolean selected;
  float time_line_value;
  float time_line_height;
  boolean redraw_me = true;

  Orga(float _x, float _y, boolean _alive) {
    x = _x;
    y = _y;
    alive = _alive;
    selected = false;
    randomGenotype();
  }
  
  Orga(float[] _genotype, int _mode) {
    for (int i = 0; i < genotype.length; i++) {
      genotype[i] = _genotype[i];
    }
    mode = _mode;
    alive = false;
    selected = false;
  }
  
  Orga(float[] _genotype, int _mode, float _tlv, float _tlh) {
    for (int i = 0; i < genotype.length; i++) {
      genotype[i] = _genotype[i];
    }
    mode = _mode;
    alive = false;
    selected = false;
    time_line_value = _tlv;
    time_line_height = _tlh;
  }
  
  public void randomGenotype() {
    for (int i = 0; i < genotype.length; i++) {
      genotype[i] = PApplet.parseFloat(PApplet.parseInt(random(0,10001))) / 10000;
    }
  }
  
  public void mutateGenotype(float mutationRate) {
    for (int i = 0; i < genotype.length; i++) {
      float nv = PApplet.parseFloat(round((genotype[i] + random(-mutationRate, mutationRate)) * 10000)) / 10000;
      if (nv < 0) nv = 1 - nv;
      nv = nv % 1;
      genotype[i] = nv;
    }
  }
  
  public void drawOrga() {
    if (redraw_me) {
      noStroke(); ellipseMode(CENTER);
      fill(1);
      ellipse(x, y, diametre_max+6, diametre_max+6);
      if (alive) {
        if (selected) {
          fill(0.03f, 0.84f, 0.99f);
          ellipse(x, y, diametre_max+4, diametre_max+4);
        }
        for (int i = (genotype.length - 1); i >= 0; i--) {
          fill(0, 0, genotype[i]);
          float d = (diametre_max / genotype.length) * (i+1);
          ellipse(x, y, d, d);
        }
      } else {
        fill(0.29f, 0.38f, 0.98f);
        ellipse(x, y, diametre_max, diametre_max);
      }
      //redraw_me = false;
    }
  }
  
  public void showParams() {
    if (redraw_me) {
      fill(1); rect(840, 110, 200, 260);
      fill(0,0,0);
      float xparam = 840;
      float yparam = 150;
      for (int i = 0; i < genotype.length; i++) {
        text("p" + i + " : " + genotype[i], xparam, yparam + 14*i);
      }
      redraw_me = false;
    }
  }
  
  public void setGenotype(float[] _genotype) {
    for (int i = 0; i < genotype.length; i++) {
      genotype[i] = _genotype[i];
    }
  }
  
  public float[] getGenotype() {
    return genotype;
  }
  
  public boolean testMouse(float _mx, float _my) {
    if (dist(_mx, _my, x, y) < (diametre_max / 2)) {
      selected = true;
      redraw_me = true;
    }
    else selected = false;
    redraw_me = true; // toujours vrai! sinon \u00e7a n'efface pas le cercle de s\u00e9lection...
    
    return selected;
  }
} 

class SoundBank {
  
  float x_start, y_start;
  float x_size = 34, y_size = 34;
  int x_length = 8, y_length = 8;
  ArrayList sounds = new ArrayList();
  int index = 0;
  int max_index = x_length * y_length - 1;
  boolean redraw_bank = true;
  
  SoundBank(float _x_start, float _y_start) {
    x_start = _x_start;
    y_start = _y_start;
  }
  
  public void drawBank() {
    if (redraw_bank) {
      fill(1); rect(x_start, y_start, x_size * x_length, y_size * y_length);
      int in_bank = sounds.size();
      int in_bank_sound = 0;
      for (int i = 0; i < y_length; i++) {
        for (int j = 0; j < x_length; j++) {
          
          float x = x_start + (j * x_size);
          float y = y_start + (i * y_size);
          stroke(0, 0, 0.6f, 1); noFill(); strokeWeight(1);
          rect(x, y, x_size, y_size);
          if (in_bank_sound < in_bank) {
            drawBankSound(in_bank_sound, x, y);
          }
          in_bank_sound ++;
        }
      }
      redraw_bank = false;
    }
    //println("banksize" + sounds.size());
  }
  
  public void drawBankSound(int _id, float _x, float _y) {
    Orga z = (Orga) sounds.get(_id);
    //ellipseMode(CORNER); noStroke(); fill(0, 1, 0.5, 1);
    noStroke(); ellipseMode(CENTER);
    
    fill(getModeColor(z.mode));
    rect(_x + 1, _y + 1, x_size - 1, y_size - 1);
    for (int i = (z.genotype.length - 1); i >= 0; i--) {
      fill(0, 0, z.genotype[i]);
      float d = ((x_size - 2) / z.genotype.length) * (i+1);
      ellipse(_x + 1 + ((x_size - 2) / 2), _y + 1 + ((y_size - 2) / 2), d, d);
    }
    //ellipse(_x + 1, _y + 1, x_size - 1, y_size - 1);
  }
  
  public void addSound(float[] _genotype, int _mode) {
    if (index <= max_index) {
      sounds.add( new Orga(_genotype, _mode));
      index++;
    } else {
      print("banque pleine");
    }
  }
  
  public void removeSound(int _id) {
    sound_bank.sounds.remove(_id);
    index--;    
  }
  
  public int mouseIn(float _mx, float _my) {
    int mouse_in = -1;
    if (     (_mx >= x_start) 
          && (_mx <= x_start + (x_size * x_length))
          && (_my >= y_start)
          && (_my <= y_start + (y_size * y_length)) )  {
       int nbx = PApplet.parseInt(( _mx - x_start ) / x_size);
       int nby = PApplet.parseInt(( _my - y_start ) / y_size);
       mouse_in = nby * y_length + nbx;
    }
    
    if (mouse_in > sounds.size() - 1) {
      mouse_in = -1;
    }
    return mouse_in;
  }
  
  public void loadBank() {
    if (sounds.size() > 0) {
      for (int i = sounds.size() - 1; i >= 0; i--) {
        sounds.remove(i);
      }
    }
    String filename = "banque_de_son.txt";
    File f = new File(dataPath(filename));
    if (f.exists()) {
      println("le fichier 'banque_de_son.txt' existe, il va \u00eatre charg\u00e9");
      String lines[] = loadStrings("./data/banque_de_son.txt");
      for (int i=0; i < lines.length; i++) {
        println(lines[i]);
        String[] val = split(lines[i], ';');
        int _mode;
        float[] _g = new float[16];

        _mode = Integer.parseInt(val[0]);
        _g[0]  = Float.parseFloat(val[1]);
        _g[1]  = Float.parseFloat(val[2]);
        _g[2]  = Float.parseFloat(val[3]);
        _g[3]  = Float.parseFloat(val[4]);
        _g[4]  = Float.parseFloat(val[5]);
        _g[5]  = Float.parseFloat(val[6]);
        _g[6]  = Float.parseFloat(val[7]);
        _g[7]  = Float.parseFloat(val[8]);
        _g[8]  = Float.parseFloat(val[9]);
        _g[9]  = Float.parseFloat(val[10]);
        _g[10] = Float.parseFloat(val[11]);
        _g[11] = Float.parseFloat(val[12]);
        _g[12] = Float.parseFloat(val[13]);
        _g[13] = Float.parseFloat(val[14]);
        _g[14] = Float.parseFloat(val[15]);
        _g[15] = Float.parseFloat(val[16]);

        sounds.add( new Orga(_g, _mode));
      }
    } else {
      println("pas de fichier 'zones.txt' \u00e0 charger");
    }
  }
  
  public void saveBank() {
    String fichier = "";
    for (int i = 0; i < sounds.size(); i++) {
      Orga z = (Orga) sounds.get(i);
      if (i > 0) fichier += " ";
      fichier += z.mode + ";"; // x + ";" + zs.y + ";" + zs.rayon;
      for (int j = 0; j < z.genotype.length; j++) { 
        if (j > 0) fichier += ";";
        fichier += z.genotype[j];
      }
    }
    String[] liste = split(fichier, ' ');
    saveStrings("./data/banque_de_son.txt", liste);
    
    println("fichier banque_de_sons.txt enregistr\u00e9");
  }
}
class Timeline {
  
  float x, y;
  float x_size = 850; // pixels
  boolean is_playing = false;
  ArrayList sounds_tl = new ArrayList();
  float playhead = 0;
  float playhead_last;
  float timeline_length = 60000; // millisec.
  float timeline_start;
  float now, last;
  boolean redraw_sounds = true;
  boolean redraw_timeline = true;

  
  Timeline(float _x, float _y) {
    x = _x;
    y = _y;  
  }
  
  public void initBySoundBank(SoundBank sb) {
    for (int i = 0; i < sb.sounds.size(); i++) {
      Orga z = (Orga) sb.sounds.get(i);
      float time = random(1001) / 1000;
      float tlh = random(40, 120);
      sounds_tl.add( new Orga(z.genotype, z.mode, time, tlh));
    }
  }
  
  public void reInit(SoundBank sb) {
    is_playing = false;
    sounds_tl = new ArrayList();
    playhead = 0;
    
    for (int i = 0; i < sb.sounds.size(); i++) {
      Orga z = (Orga) sb.sounds.get(i);
      float time = random(1001) / 1000;
      float tlh = random(40, 120);
      sounds_tl.add( new Orga(z.genotype, z.mode, time, tlh));
    }
    redraw_sounds = true;
  }
  
  public void doActions() {
    if (redraw_timeline) {
      drawTimeline();
      redraw_timeline = false;
    }
    if (is_playing) {
      drawPlay();
      redraw_timeline = true;
    }
    if (redraw_sounds) {
      drawSounds();
      redraw_sounds = false;
    }
  }
  
  public void drawTimeline() {
      fill(0, 0, 0.6f, 1);
      noStroke();
      rectMode(CORNER);
      rect (x, y, x_size, 20);
  }
  
   public void drawPlay() {
    fill(0, 1, 0.6f, 1);
    noStroke();
    rectMode(CORNER);
    rect (x, y, x_size * playhead, 20);
  }
  
  public void drawSounds() {
    fill(1); rect(x - 30, y - 140, x_size + 60, 140);
    for (int i = 0; i < sounds_tl.size(); i++) {
      //float yh = random(40, 120);
      Orga z = (Orga) sounds_tl.get(i);
      float xs = z.time_line_value * x_size;
      float yh = z.time_line_height;
      //stroke(0, 0, 0, 1); noFill();
      //line(x + xs, y , x + xs, y - 50); 
      noStroke(); fill(getModeColor(z.mode));
      beginShape();
        vertex(x + xs - 4, y - yh);
        vertex(x + xs + 4, y - yh);
        vertex(x + xs, y);
      endShape(CLOSE);
      noStroke(); ellipseMode(CENTER);
    
      fill(getModeColor(z.mode));
      float xx = x + xs;
      float yy = y - yh;
      ellipse(xx, yy, 36, 36); //rect(_x + 1, _y + 1, x_size - 1, y_size - 1);
      for (int j = (z.genotype.length - 1); j >= 0; j--) {
        fill(0, 0, z.genotype[j]);
        float d = (32 / z.genotype.length) * (j+1);
        ellipse(xx, yy, d, d);
      }
    }
  }
  
  public void start(float ms) {
    is_playing = true;
    timeline_start = ms;
    last = ms;
    now = ms;
    playhead = 0;
    playhead_last = 0;
  }
  
  public void play(float ms) {
    now = ms - last;
    playhead = playhead + (now / timeline_length);
    //println("now " + now + " last " + last + " ms " + ms + " playhead " + playhead + " playhead_last " + playhead_last);
    playSounds(playhead_last, playhead);
    playhead_last = playhead;
    last = ms;
  }
  
  public boolean endReached() {
    boolean er = false;
    if (playhead > 1) { // STOP
      is_playing = false;
      playhead = 0;
      playhead_last = 0;
      er = true;
    }
    return er;
  }
  
  public void playSounds(float ph_start, float ph_end) {
    for (int i = 0; i < sounds_tl.size(); i++) {
      Orga z = (Orga) sounds_tl.get(i);
      if ( (z.time_line_value >= ph_start) &&  (z.time_line_value < ph_end) ) {
        sendGenotypeOsc(z.genotype, z.mode);
      }
    }
  }
  // TODO
  
  public void sendGenotypeOsc(float[] _genotype, int mode) {
  sendModeOsc(mode);
  OscMessage message = new OscMessage("/params");
  for (int i = 0; i < max_params; i++) {
    message.add(_genotype[i]); 
  }
  osca.send(message, loc); 
  }

  public void sendModeOsc(int mode) {
    OscMessage message = new OscMessage("/mode");
    message.add(mode); 
    osca.send(message, loc); 
  }
  
  public void saveTimeline() {
    String fichier = "";
    for (int i = 0; i < sounds_tl.size(); i++) {
      Orga z = (Orga) sounds_tl.get(i);
      if (i > 0) fichier += " ";
      fichier += z.time_line_value + ";";
      fichier += z.mode + ";"; // x + ";" + zs.y + ";" + zs.rayon;
      for (int j = 0; j < z.genotype.length; j++) { 
        if (j > 0) fichier += ";";
        fichier += z.genotype[j];
      }
    }
    String[] liste = split(fichier, ' ');
    saveStrings("./data/timeline.txt", liste);
    
    println("fichier timeline.txt enregistr\u00e9");
  }
  
}
public void showHelp() { // TODO !
  fill(0, 0, 1, 1); stroke(0, 0, 0, 1); strokeWeight(1);
  rect (100, 100, 550, 550);
  String notes = "";
  notes += "\n";
  notes += "clic gauche : choisir comme parent A\n\n";
  notes += "clic droit : choisir comme parent B\n\n";
  notes += "s : sauver une copie d'\u00e9cran\n\n";
  notes += "c : croiser les parents et cr\u00e9er une nouvelle population\n\n";
  notes += "h : afficher / cacher l'aide\n\n";
  notes += "x : sauver la banque de sons (banque_de_sons.txt)\n\n";
  notes += "y : charger la derni\u00e8re banque de sons enregistr\u00e9e\n\n";
  notes += "l : jouer la timeline (2 minutes)\n\n";
  notes += "n : r\u00e9initialiser la timeline\n\n";
  notes += "t : enregistrer la timeline (timeline.txt)\n\n";
  notes += "r : remplacer par un g\u00e9notype random\n\n";
  notes += "m : muter le g\u00e9notype (5%)\n\n";
  notes += "o : jouer imm\u00e9diatement ce g\u00e9notype\n\n";
  notes += "d : enlever ce son de la banque\n\n";
  notes += "a : jouer imm\u00e9diatement le parent A\n\n";
  notes += "b : jouer imm\u00e9diatement le parent B\n\n";
  notes += "k : ajouter \u00e0 la banque\n\n";
  notes += "b : jouer imm\u00e9diatement le parent B\n\n";
  notes += "b : jouer imm\u00e9diatement le parent B\n\n";
  notes += "0...9 : changer le mode de cr\u00e9atino sonore\n\n\n\n";

  
  fill(0, 0, 0, 1); stroke(0, 0, 0, 1);
  text(notes, 120, 115, 550, 550);
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "ga" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
